// exbsa, v1.2.1 2014/02/26
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts data from newer BSS-Arc (*.bsa) archives.

#include "as-util.h"
#include <list>

// #define BSSARC_VERSION 1
#define BSSARC_VERSION 2

#pragma pack(1)
struct BSSARCHDR {
  uint8_t  signature[8]; // "BSS-Arc"
  uint16_t version;      // ??
  uint16_t entry_count;
  uint32_t toc_offset;
};

struct BSSARCENTRY {
#if BSSARC_VERSION >= 2
  uint32_t filename_offset;
#else
  char     filename[32];
#endif
  uint32_t offset;
  uint32_t length;
};

struct BSSCOMPHDR {
  uint8_t  signature[16]; // "BSS-Composition"
  uint8_t  unknown1;
  uint32_t entry_count;
  uint8_t  unknown2;
  uint16_t base_width;
  uint16_t base_height;
  uint8_t  unknown3[6];
};

struct BSSGRAPHDR {
  uint8_t  signature[16]; // "BSS-Graphics"
  uint8_t  unknown1[16];
  uint16_t offset_x;
  uint16_t offset_y;
  uint8_t  unknown2[4];
  uint16_t width;
  uint16_t height;
  uint8_t  unknown3;
  uint8_t  frame_count;
  uint8_t  unknown4[2];
  uint8_t  type; // 0 = 32-bit, 1 = 24-bit, 2 = 8-bit?
  uint8_t  compression_type;
  uint8_t  unknown5[4];
  uint32_t length;
  uint8_t  unknown6[6];  
};
#pragma pack()

static const uint8_t COMPRESSION_TYPE_NONE = 0;
static const uint8_t COMPRESSION_TYPE_RLE  = 1;
static const uint8_t COMPRESSION_TYPE_LZ   = 2;

uint8_t* unlz(uint8_t* buff,
              uint8_t* out_buff,
              uint32_t pixel_bytes)
{
  for (uint32_t color = 0; color < pixel_bytes; color++) {
    uint8_t  control        = *buff++;
    uint32_t compressed_len = *(uint32_t*)buff;
    buff += 4;

    auto end   = buff + compressed_len - 5;
    auto out_p = out_buff + color;

    while (buff < end) {
      uint8_t c = *buff++;

      if (c == control ) {
        uint32_t p = *buff++;

        if (p != control) {
          uint32_t n = *buff++;

          if (p > control) {
            p--;
          }

          p *= pixel_bytes;

          while (n--) {
            *out_p = *(out_p - p);
            out_p += pixel_bytes;
          }
        } else {
          *out_p = c;
          out_p += pixel_bytes;
        }
      } else {
        *out_p = c;
        out_p += pixel_bytes;
      }
    }

    auto out_end = out_p;
    out_p = out_buff + color + pixel_bytes;

    while (out_p < out_end) {
      *out_p += *(out_p - pixel_bytes);
      out_p += pixel_bytes;
    }
  }

  return buff;
}

uint8_t* unrle(uint8_t* buff,
               uint8_t* out_buff,
               uint32_t pixel_bytes)
{
  for (uint32_t color = 0; color < pixel_bytes; color++) {
    uint32_t compressed_len = *(uint32_t*) buff;
    buff += 4;

    auto end   = buff + compressed_len;
    auto out_p = out_buff + color;

    while (buff < end) {
      uint32_t n = *buff++;

      if (n & 0x80) {
        n = 1 - (char)n;

        while (n--) {
          *out_p = *buff;
          out_p += pixel_bytes;
        }

        buff++;
      } else {
        n++;

        while (n--) {
          *out_p = *buff++;
          out_p += pixel_bytes;
        }
      }
    }
  }

  return buff;
}

bool process_graphic(const string& prefix, 
                     uint8_t*      buff)
{
  auto hdr  = (BSSGRAPHDR*) buff;
  auto data = (uint8_t*) (hdr + 1);

  if (memcmp(hdr->signature, "BSS-Graphics", 12)) {
    return false;
  }

  uint32_t pixel_bytes = 0;
  
  switch (hdr->type) {
  case 0:
    pixel_bytes = 4;
    break;
  case 1:
    pixel_bytes = 3;
    break;
  case 2:
    pixel_bytes = 1;
    break;
  default:
    return false;
  };

  as::image_t image(hdr->width, hdr->height, pixel_bytes);

  uint8_t* after = NULL;

  switch (hdr->compression_type) {
  case COMPRESSION_TYPE_NONE:
    memcpy(image.buffer(), data, image.length());
    after = data + image.length();
    break;

  case COMPRESSION_TYPE_RLE:
    after = unrle(data, image.buffer(), pixel_bytes);
    break;

  case COMPRESSION_TYPE_LZ:
    after = unlz(data, image.buffer(), pixel_bytes);
    break;
  }

  if (pixel_bytes == 1) {
    image.set_palette(after, 256);
  }

  string filename = prefix;

  if (hdr->offset_x || hdr->offset_y) {
    filename += as::stringf("+x%dy%d", hdr->offset_x, hdr->offset_y);
  }

  filename += ".bmp";

  image.write(filename);

  return true;
}

bool process_composition(const string&  prefix, 
                         uint8_t* buff) 
{
  auto hdr  = (BSSCOMPHDR*) buff;
  auto data = (uint8_t*) (hdr + 1);

  if (memcmp(hdr->signature, "BSS-Composition", 15)) {
    return false;
  }

  bool has_error = false;

  for (uint32_t i = 0; i < hdr->entry_count; i++) {
    if (!process_graphic(prefix + as::stringf("+%03d", i), data)) {
      fprintf(stderr, "%s: failed to extract entry %d\n", prefix.c_str(), i);
      has_error = true;
    }

    auto grahdr = (BSSGRAPHDR*) data;
    data += sizeof(*grahdr) + grahdr->length;
  }

  return !has_error;
}

class directories_t {
public:
  void push(const string& d) {
    dirnames.push_back(d);
  }

  void pop(void) {
    dirnames.pop_back();
  }

  string get_path(const string& f) {
    string s;

    for (dirnames_t::iterator i = dirnames.begin();
         i != dirnames.end();
         ++i)
    {
      s += *i + "/";
    }

    return s + f;
  }

private:
  typedef std::list<string> dirnames_t;
  dirnames_t dirnames;
};

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "exbsa v1.2.1 by asmodean\n\n");
    fprintf(stderr, "usage: %s <input.bsa>\n", argv[0]);
    return -1;
  }

  string in_filename = argv[1];

  int fd = as::open_or_die(in_filename, O_RDONLY | O_BINARY);

  BSSARCHDR hdr;
  read(fd, &hdr, sizeof(hdr));

  auto entries = new BSSARCENTRY[hdr.entry_count];
  lseek(fd, hdr.toc_offset, SEEK_SET);
  read(fd, entries, sizeof(BSSARCENTRY) * hdr.entry_count);

#if BSSARC_VERSION >= 2
  uint32_t filenames_len  = as::get_file_size(fd) - tell(fd);
  auto     filenames_buff = new char[filenames_len];
  read(fd, filenames_buff, filenames_len);
#endif

  directories_t dirs;

  for (uint32_t i = 0; i < hdr.entry_count; i++) {
#if BSSARC_VERSION >= 2
    char* filename = filenames_buff + entries[i].filename_offset;
#else
    char* filename = entries[i].filename;
#endif

    switch (filename[0]) {
    case '>':
      dirs.push(filename + 1);
      break;

    case '<':
      dirs.pop();
      break;

    default:
      {
        uint32_t len  = entries[i].length;
        auto     buff = new uint8_t[len];
        lseek(fd, entries[i].offset, SEEK_SET);
        read(fd, buff, len);

        string full_filename = dirs.get_path(filename);

        as::make_path(full_filename);          

        if (!process_graphic(as::get_file_prefix(full_filename), buff) &&
            !process_composition(as::get_file_prefix(full_filename), buff))
        {
          as::write_file(full_filename, buff, len);
        }

        delete [] buff;
      }
    }
  }

#if BSSARC_VERSION >= 2
  delete [] filenames_buff;
#endif

  delete [] entries;

  return 0;
}
